"""Generate a status line for swaybar."""

import toml

from argparse import ArgumentParser
from pathlib import Path

from .config import config
from .env import bin_name, config_home, environ_path, environ_paths
from .logging import logger, configure as configure_logging
from .loop import start
from .modules import Modules


def parse_args():
    parser = ArgumentParser(description=__doc__)

    parser.add_argument(
        "-c",
        "--config-file",
        metavar="FILE",
        type=Path,
        help="override configuration file",
    )

    parser.add_argument(
        "-C",
        "--config-dir",
        metavar="DIRECTORY",
        type=Path,
        help="override configuration directory",
    )

    parser.add_argument(
        "-I",
        "--include",
        action="append",
        metavar="DIRECTORY",
        type=Path,
        help="include additional modules package",
    )

    parser.add_argument(
        "-i",
        "--interval",
        type=float,
        metavar="SECONDS",
        help="override default update interval",
    )

    parser.add_argument(
        "--no-click-events",
        dest="click_events",
        action="store_false",
        help="disable click events",
    )

    parser.add_argument(
        "-L",
        "--log-level",
        metavar="LEVEL",
        choices=["debug", "info", "warning", "error", "critical"],
        help="override default logging minimum severity level",
    )

    parser.add_argument(
        "-l",
        "--log-file",
        metavar="FILE",
        type=Path,
        help="output logging to %(metavar)s",
    )

    parser.add_argument(
        "--syslog",
        action="store_true",
        help="output logging to syslog",
    )

    return parser.parse_args()


def parse_config(args):
    config_dir = args.config_dir or environ_path(
        "SWAYSTATUS_CONFIG_DIR", config_home / bin_name
    )
    config_file = args.config_file or environ_path(
        "SWAYSTATUS_CONFIG_FILE", config_dir / "config.toml"
    )

    if config_file.is_file():
        config.update(toml.loads(open(config_file).read()))

    config["include"] = (
        (args.include or [])
        + [config_dir / "modules"]
        + [Path(d).expanduser() for d in config.get("include", [])]
        + environ_paths("SWAYSTATUS_MODULE_PATH")
    )

    if args.interval:
        config["interval"] = args.interval

    if not args.click_events:
        config["click_events"] = False

    return config


def deep_merge_dicts(first, second):
    """
    Recursively merge the second dictionary into the first.
    """
    result = first.copy()
    for key, value in second.items():
        if key in result and isinstance(result[key], dict) and isinstance(value, dict):
            result[key] = deep_merge_dicts(result[key], value)
        else:
            result[key] = value
    return result


def load_elements(order, include, settings):
    elements = []
    modules = Modules(include)

    for key in order:
        try:
            name, instance = key.split(":", maxsplit=1)
        except ValueError:
            name, instance = key, None

        module = modules.find(name)

        kwargs = deep_merge_dicts(
            settings.get(name, {}),
            settings.get(key, {}),
        )

        kwargs.update({"name": name, "instance": instance})

        logger.info(f"Loaded module from file: {module.__file__}")
        logger.debug(f"Initializing module: {kwargs!r}")

        elements.append(module.Element(**kwargs))

    return elements


def main():
    args = parse_args()

    configure_logging(
        level=args.log_level,
        file=args.log_file,
        syslog=args.syslog,
    )

    config = parse_config(args)
    logger.debug(f"Using configuration: {config!r}")

    elements = load_elements(
        config["order"],
        config["include"],
        config["settings"],
    )

    try:
        start(
            elements,
            config["interval"],
            config["click_events"],
        )
    except Exception:
        logger.exception("Unhandled exception in main loop")
        return 1

    return 0



from logging import getLogger, basicConfig, Formatter, StreamHandler, FileHandler

from .env import bin_name

logger = getLogger(bin_name)


def create_formatter(named=True, timestamped=True):
    fmt = ""

    if timestamped:
        fmt += "%(asctime)s: "

    fmt += "%(levelname)s: "

    if named:
        fmt += "%(name)s: "

    fmt += "%(message)s"

    return Formatter(fmt)


def configure(level=None, file=None, syslog=False):
    handlers = []

    stream_handler = StreamHandler()
    stream_handler.setFormatter(create_formatter())
    handlers.append(stream_handler)

    if file:
        file_handler = FileHandler(file)
        file_handler.setFormatter(create_formatter())
        handlers.append(file_handler)

    if syslog:
        from logging.handlers import SysLogHandler

        syslog_handler = SysLogHandler(address="/dev/log")
        syslog_handler.setFormatter(create_formatter(named=False, timestamped=False))
        syslog_handler.ident = f"{bin_name}: "

        handlers.append(syslog_handler)

    if level and isinstance(level, str):
        level = level.upper()

    basicConfig(level=level, handlers=handlers)



import sys

from importlib import import_module, metadata
from importlib.util import spec_from_file_location, module_from_spec
from pathlib import Path
from uuid import uuid4


def unique_package_name():
    return str(uuid4()).replace("-", "")


class Modules:
    def __init__(self, include):
        self._packages = []
        self._cached_modules = {}

        for i, modules_dir in enumerate(include):
            if (init_file := Path(modules_dir).expanduser() / "__init__.py").is_file():
                package_name = unique_package_name()
                if spec := spec_from_file_location(package_name, init_file):
                    package = module_from_spec(spec)
                    sys.modules[package_name] = package
                    if spec.loader:
                        spec.loader.exec_module(package)
                        self._packages.append(package_name)

        for entry_point in metadata.entry_points(group="swaystatus.modules"):
            self._packages.append(entry_point.load().__name__)

    def find(self, name):
        if name not in self._cached_modules:
            for package in self._packages:
                try:
                    self._cached_modules[name] = import_module(f"{package}.{name}")
                    break
                except ModuleNotFoundError:
                    continue
            else:
                raise ModuleNotFoundError(f"Module not found in any package: {name}")

        return self._cached_modules[name]



import json
import time

from signal import SIGSTOP, SIGCONT


def send_line(line):
    print(line, flush=True)


class Updater:
    def __init__(self, elements, interval, click_events):
        super().__init__()

        self.elements = elements
        self.element_timers = []
        for element in self.elements:
            self.element_timers.append([0] * len(element.intervals))

        self.interval = interval
        self.time_before = time.perf_counter()

        self._header = {
            "version": 1,
            "stop_signal": SIGSTOP,
            "cont_signal": SIGCONT,
            "click_events": click_events,
        }
        self._body_start = "[[]"
        self._body_item = ",{}"

        self._running = False

    def update(self):
        """
        Prompt every element for any updates to the status bar.

        It does this by giving each element a turn at appending blocks to an
        `output` list that it passes as the first argument to the element's
        `on_update` method. This is done in the order that the elements were
        given to the updater at initialization.

        It also determines if any element has intervals that have come due and
        should be triggered by calling the element's `on_interval` method with
        the options given during `set_interval`.

        After all this has been done, a body item, as described in the "BODY"
        section of swaybar-protocol(7), is constructed and sent immediately to
        stdout.
        """

        time_now = time.perf_counter()
        self.seconds_elapsed = time_now - self.time_before
        self.time_before = time_now

        output = []

        for element_index, element in enumerate(self.elements):
            timers = self.element_timers[element_index]
            for interval_index, timer in enumerate(timers):
                timer += self.seconds_elapsed
                interval, options = element.intervals[interval_index]
                if timer >= interval:
                    element.on_interval(options)
                    timers[interval_index] = 0
                else:
                    timers[interval_index] = timer
            element.on_update(output)

        send_line(self._body_item.format(json.dumps(output)))

    def running(self):
        return self._running

    def stop(self):
        self._running = False

    def start(self):
        send_line(json.dumps(self._header))
        send_line(self._body_start)

        self._running = True

        while self.running():
            self.update()
            time.sleep(self.interval)



from subprocess import Popen, PIPE
from threading import Thread


class PopenStreamHandler(Popen):
    """
    Just like `Popen`, but handle stdout and stderr output in dedicated
    threads.
    """

    def __init__(self, stdout_handler, stderr_handler, *args, **kwargs):
        def proxy(context, handler):
            with context as lines:
                for line in lines:
                    handler(line)

        kwargs["stdout"] = PIPE
        kwargs["stderr"] = PIPE

        super().__init__(*args, **kwargs)

        Thread(target=proxy, args=[self.stdout, stdout_handler]).start()
        Thread(target=proxy, args=[self.stderr, stderr_handler]).start()



import os
import sys

from pathlib import Path

bin_name = os.path.basename(sys.argv[0])
config_home = Path(os.environ.get("XDG_CONFIG_HOME", "~/.config")).expanduser()


def environ_path(name, default=None):
    value = os.environ.get(name, default)
    return Path(value).expanduser() if value else default


def environ_paths(name, default=None):
    value = os.environ.get(name)
    parts = value.split(":") if value else (default or [])
    return [Path(p).expanduser() for p in parts]
